/*
    x_glx.c - X11/glx graphic output
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/glx.h>
#include "x_glx.h"


#ifndef GLX_VERSION_1_3
#define USE_OLD_GLX12
#warning glX headers version < 1.3
#endif


#ifdef DEBUG_GLX
#define GLDEBUG(s...)	fprintf(stderr, "x_glx: " s)
#else
#define GLDEBUG(s...)
#endif

struct x_display_t {
	Display *display;
	int initial_auto_repeat;
#ifndef USE_OLD_GLX12
#define DISPLAY_HAS_GLX13	1
	int flags;
#endif
};

struct x_window_t {
	Display *display;
	Window window;

	GLXContext ctx;
#ifndef USE_OLD_GLX12
	int flags;
	
	GLXWindow glxw;
#endif
};

// ****************************************************************
// Open and close X display connection
// ****************************************************************
x_display *x_open_display(void)
{
	XKeyboardState kbd;
	x_display *r;
	int major, minor;
	
	r = (x_display *)malloc(sizeof(*r));
	if (!r) return(NULL);

	r->display = XOpenDisplay(NULL);
	if (!r->display) {
		GLDEBUG("Could not open display\n");
		free(r);
		return(NULL);
	}

	if (!glXQueryExtension(r->display, NULL, NULL)) {
		GLDEBUG("X Server does not support the GLX extension\n");
		XCloseDisplay(r->display);
		free(r);
		return(NULL);
	}
	
	if (!glXQueryVersion(r->display, &major, &minor)) {
		GLDEBUG("X Server does not report the version of its "
			"GLX extension\n");
		XCloseDisplay(r->display);
		free(r);
		return(NULL);
	}

	
#ifdef USE_OLD_GLX12
	GLDEBUG("glx 1.2: server-side glx %d.%d detected.\n", major, minor);
#else
	GLDEBUG("glx 1.3+: server-side glx %d.%d detected.\n", major, minor);
	r->flags = 0;
	
	if ((major > 1) || ((major == 1) && (minor >= 3))) {
		r->flags |= DISPLAY_HAS_GLX13;
	} else {
		GLDEBUG("Using legacy GLX 1.2 protocol.\n");
	}
#endif

	XGetKeyboardControl(r->display, &kbd);
	r->initial_auto_repeat = (kbd.global_auto_repeat == AutoRepeatModeOn);

	GLDEBUG("Autorepeat is %s\n",
		r->initial_auto_repeat ? "enabled" : "disabled");
	
	return(r);
}


void x_close_display(x_display *r)
{
	if (r->initial_auto_repeat) XAutoRepeatOn(r->display);
	else XAutoRepeatOff(r->display);

	GLDEBUG("Closing display\n");

	XSync(r->display, False);
	XCloseDisplay(r->display);
	free(r);
}


// ****************************************************************
// Create and destroy X window
// ****************************************************************
x_window  *x_create_window(x_display *r,
	int x, int y, int width, int height, int flags)
{
	x_window *u;
	unsigned long attr_mask;
	XSetWindowAttributes attr;
	XVisualInfo *info;
#ifndef USE_OLD_GLX12
	GLXFBConfig *cfgs;
	GLXFBConfig cfg = (GLXFBConfig)-1;
#endif

	u = (x_window *)malloc(sizeof(*u));
	if (!u) return(NULL);

	u->display = r->display;

#ifndef USE_OLD_GLX12
	u->flags = r->flags;
	
	if (r->flags & DISPLAY_HAS_GLX13) {
		int n;
		int attrl[] = {
			GLX_X_RENDERABLE, True,
			GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
			GLX_RENDER_TYPE, GLX_RGBA_BIT,
			GLX_DOUBLEBUFFER, True,
			GLX_RED_SIZE, 4,
			GLX_GREEN_SIZE, 4,
			GLX_BLUE_SIZE, 4,
			GLX_DEPTH_SIZE, 12,
			None };
		//GLX_CONFIG_CAVEAT
		//  ==GLX_NONE/GLX_SLOW_CONFIG/GLX_NON_CONFORMANT_CONFIG
		//GLX_ACCUM_x_SIZE, GLX_STENCIL_SIZE, GLX_AUX_BUFFERS
		//GLX_TRANSPARENT_TYPE should be GLX_NONE

		cfgs = glXChooseFBConfig(u->display, DefaultScreen(u->display),
			attrl, &n);
		if (!cfgs) {
			free(u);
			return(NULL);
		}

		// Should be what glx considers to be the best. Works for now.
		cfg = cfgs[0];
		XFree(cfgs);

		info = glXGetVisualFromFBConfig(u->display, cfg);
		if (!info) {
			free(u);
			return(NULL);
		}

		GLDEBUG("glx 1.3+: picked FBConfig %p (visualid 0x%x) "
			"among %d available\n",
			cfg, (int)info->visualid, n);

		u->ctx = glXCreateNewContext(u->display, cfg, GLX_RGBA_TYPE,
			NULL, True);
		if (!u->ctx) {
			XFree(info);
			free(u);
			return(NULL);
		}
	} else
#endif
	{
		int attrl12[] = {
			GLX_RGBA, GLX_DOUBLEBUFFER,
			GLX_RED_SIZE, 4,
			GLX_GREEN_SIZE, 4,
			GLX_BLUE_SIZE, 4,
			GLX_DEPTH_SIZE, 12,
			0 };

		info = glXChooseVisual(u->display, DefaultScreen(u->display),
			attrl12);
		if (!info) {
			free(u);
			return(NULL);
		}

		GLDEBUG("glx 1.2: picked visual id 0x%x\n",
			(int)info->visualid);

		u->ctx = glXCreateContext(u->display, info, NULL, True);
		if (!u->ctx) {
			XFree(info);
			free(u);
			return(NULL);
		}
	}

	GLDEBUG("Using (%sdirect rendering) context %p\n",
		glXIsDirect(u->display, u->ctx) ? "" : "in", u->ctx);

	attr_mask = CWBackPixmap|CWBackingStore|CWOverrideRedirect|CWSaveUnder
		|CWEventMask|CWColormap;
	attr.background_pixmap = None;
	attr.backing_store = NotUseful;
	attr.override_redirect = (flags & X_WINDOW_BYPASS_WM) != 0;
	attr.save_under = False;
	attr.event_mask = KeyPressMask|KeyReleaseMask|ButtonPressMask
		|ButtonReleaseMask|PointerMotionMask|KeymapStateMask
		|EnterWindowMask|LeaveWindowMask|FocusChangeMask
		|ExposureMask|StructureNotifyMask;
	attr.colormap = XCreateColormap(u->display,
		RootWindow(u->display, DefaultScreen(u->display)),
		info->visual, AllocNone);

	u->window = XCreateWindow(u->display,
		RootWindow(u->display, DefaultScreen(u->display)),
		x, y, width, height, 0, info->depth, InputOutput,
		info->visual, attr_mask, &attr);

	if (u->window == None) {
		XFree(info);
		free(u);
		return(NULL);
	}

	XFree(info);
#ifndef USE_OLD_GLX12
	if (r->flags & DISPLAY_HAS_GLX13) {
		// Some drivers segfault on NULL attribute list
		int glxwin_attr[] = { None };

		u->glxw = glXCreateWindow(u->display, cfg, u->window,
			glxwin_attr);

		if (u->glxw == None) {
			XDestroyWindow(u->display, u->window);
			glXDestroyContext(u->display, u->ctx);
			free(u);
			return(NULL);
		}
		
		if (!glXMakeContextCurrent(u->display, u->glxw, u->glxw,
		u->ctx)) {
			XDestroyWindow(u->display, u->window);
			glXDestroyContext(u->display, u->ctx);
			glXDestroyWindow(u->display, u->glxw);
			free(u);
			return(NULL);
		}
	} else
#endif
	{
		if (!glXMakeCurrent(u->display, u->window, u->ctx)) {
			XDestroyWindow(u->display, u->window);
			glXDestroyContext(u->display, u->ctx);
			free(u);
			return(NULL);
		}
	}
	
	XMapRaised(u->display, u->window);
	XSync(u->display, False);

	GLDEBUG("Window 0x%x ready\n", (int)u->window);

	return(u);
}

// ****************************************************************
// 
// ****************************************************************
void x_destroy_window(x_window *u)
{

	GLDEBUG("Destroying window 0x%x\n", (int)u->window);

#ifndef USE_OLD_GLX12
	if (u->flags & DISPLAY_HAS_GLX13)
		glXDestroyWindow(u->display, u->glxw);
#endif
	glXDestroyContext(u->display, u->ctx);
	XDestroyWindow(u->display, u->window);
	free(u);
}


// ****************************************************************
// 
// ****************************************************************
void x_set_window_name(x_window *u, char *name, char *icon_name)
{
#ifdef X_HAVE_UTF8_STRING
	Xutf8SetWMProperties(u->display, u->window, name, icon_name,
		NULL, 0, NULL, NULL, NULL);
#else
	XmbSetWMProperties(u->display, u->window, name, icon_name,
		NULL, 0, NULL, NULL, NULL);
#endif
}

// ****************************************************************
// 
// ****************************************************************
void x_move_resize_window(x_window *u, int x, int y, int w, int h)
{
	if ((x >= 0) && (y >= 0) && (w >= 0) && (h >= 0)) {
		XMoveResizeWindow(u->display, u->window, x, y, w, h);
		return;
	}

	if ((w >= 0) && (h >= 0)) {
		XResizeWindow(u->display, u->window, w, h);
		return;
	}

	XMoveWindow(u->display, u->window, x, y);
}

// ****************************************************************
// 
// ****************************************************************
void x_set_window_size_range(x_window *u, int w1, int h1, int w2, int h2)
{
	XSizeHints sz_hints;

	sz_hints.flags = PMinSize|PMaxSize;
	sz_hints.min_width  = w1; sz_hints.max_width  = w2;
	sz_hints.min_height = h1; sz_hints.max_height = h2;

	XSetWMProperties(u->display, u->window, NULL, NULL,
		NULL, 0, &sz_hints, NULL, NULL);
}

// ****************************************************************
// 
// ****************************************************************
int  x_get_window_size(x_window *u, int *width, int *height)
{
	XWindowAttributes attr;

	if (!XGetWindowAttributes(u->display, u->window, &attr)) return(-1);

	if (width) *width = attr.width;
	if (height) *height = attr.height;
	return(0);
}

// ****************************************************************
// 
// ****************************************************************
void x_swap_buffers(x_window *u)
{
#ifndef USE_OLD_GLX12
	if (u->flags & DISPLAY_HAS_GLX13)
		glXSwapBuffers(u->display, u->glxw);
	else
#endif
		glXSwapBuffers(u->display, u->window);
}


// ****************************************************************
// 
// ****************************************************************
void x_sync(x_display *r)
{
	XSync(r->display, False);
}

void x_flush(x_display *r)
{
	XFlush(r->display);
}

// ****************************************************************
// 
// ****************************************************************
int  x_get_input(x_display *r, int wait)
{
	XEvent e;
	
	while(1) {
		if (!wait) {
			if (XPending(r->display) == 0) return(0);
		}
		
		XNextEvent(r->display, &e);
		
		if (e.type == KeyPress) {
			return(e.xkey.keycode);
		}
		
		if (e.type == ButtonPress) {
			return(e.xbutton.button);
		}
		
		if ((e.type == Expose) && (e.xexpose.count == 0)) return(-1);
	}
}


